import ast
import contextlib
import inspect
from collections import defaultdict

from executing import Source

from ._rewrite_code import ChangeRecorder
from ._rewrite_code import end_of
from ._rewrite_code import start_of

# sentinels
missing_value = object()
undefined = object()

snapshots = {}

# TODO: optimize later
_force_replace = True

_ignore_value = False
_active = False


@contextlib.contextmanager
def snapshots_disabled():
    global snapshots
    current = snapshots
    snapshots = {}
    yield
    snapshots = current


def fix_snapshots(reasons):
    for snapshot in snapshots.values():
        if snapshot._reason in reasons or "all" in reasons:
            snapshot._change()


def snapshot_stat():
    stat = defaultdict(int)
    for snapshot in snapshots.values():
        stat[snapshot._reason] += 1

    return stat


def snapshot(obj=missing_value):

    if not _active:
        if obj is missing_value:
            raise AssertionError(
                "your snapshot is missing a value run pytest with --update-snapshots=new"
            )
        else:
            return obj

    global _snapshot_id

    frame = inspect.currentframe().f_back
    expr = Source.executing(frame)

    key = expr.node
    key = id(frame.f_code), frame.f_lasti

    if key not in snapshots:
        node = expr.node
        if node is None:
            # we can run without knowing of the calling expression but we will not be able to fix code
            snapshots[key] = Snapshot(obj, None)
        else:
            assert isinstance(node.func, ast.Name)
            assert node.func.id == "snapshot"
            snapshots[key] = Snapshot(obj, expr)

    return snapshots[key]


class Snapshot:
    def __init__(self, value, expr):

        self._expr = expr
        self._new_value = undefined

        self._current_value = value
        self._reason = None

    def __repr__(self):
        if self._current_value is not missing_value:
            return repr(self._current_value)
        else:
            return repr(self._new_value)

    def _set_argument(self, o, reason):
        if o == self._new_value:
            return
        self._reason = reason
        self._new_value = o

    def _change(self):
        assert self._expr is not None

        change = ChangeRecorder.current.new_change()

        tokens = list(self._expr.source.asttokens().get_tokens(self._expr.node))
        assert tokens[0].string == "snapshot"
        assert tokens[1].string == "("
        assert tokens[-1].string == ")"

        change.set_tags("inline_snapshot", self._reason)

        change.replace(
            (end_of(tokens[1]), start_of(tokens[-1])),
            repr(self._new_value),
            filename=self._expr.source.filename,
        )

    def __eq__(self, o):

        if self._current_value == missing_value:
            # x==snapshot()
            # assert "foo" == snapshot()
            # always true because we want to execute the rest of the code
            if self._new_value is undefined:
                # first call
                self._set_argument(o, "new")
                return True
            else:
                # second call
                return o == self._new_value

        correct = self._current_value == o

        if not correct:
            self._set_argument(o, "failing")
        else:
            if _force_replace:
                self._set_argument(o, "force")

        if _ignore_value:
            return True

        return self._current_value == o



from . import _inline_snapshot
from ._rewrite_code import ChangeRecorder


def pytest_addoption(parser):
    group = parser.getgroup("inline-snapshot")
    group.addoption(
        "--update-snapshots",
        action="store",
        dest="inline_snapshot_kind",
        default="none",
        choices=("all", "failing", "new", "none"),
        help="update snapshot arguments in source code:\n [force] incorrect and correct snapshots\n [failing] incorrect snapshots\n [new] snapshots without a value",
    )

    # TODO --snapshots-value-required ... fail if there are pending updates


def pytest_configure(config):

    if config.option.inline_snapshot_kind == "failing":
        _inline_snapshot._ignore_value = True

    if config.option.inline_snapshot_kind != "none":
        _inline_snapshot._active = True

        import sys

        # hack to disable the assertion rewriting
        # I found no other way because the hook gets installed early
        sys.meta_path = [
            e for e in sys.meta_path if type(e).__name__ != "AssertionRewritingHook"
        ]


def pytest_terminal_summary(terminalreporter, exitstatus, config):
    recorder = ChangeRecorder.current

    terminalreporter.section("inline snapshot")
    if exitstatus != 0:
        failing = sum(
            1
            for snapshot in _inline_snapshot.snapshots.values()
            if snapshot._reason == "failing"
        )

        if failing:
            terminalreporter.write(
                f"{failing} snapshots are causing problems (--update-snapshots=failing)\n"
            )

    new = sum(
        1
        for snapshot in _inline_snapshot.snapshots.values()
        if snapshot._reason == "new"
    )

    if new:
        terminalreporter.write(
            f"{new} snapshots are missing values (--update-snapshots=new)\n"
        )

    fix_reason = config.option.inline_snapshot_kind

    if fix_reason != "none":

        new = 0
        failing = 0

        for snapshot in _inline_snapshot.snapshots.values():
            if snapshot._reason == "new" == fix_reason:
                new += 1
                snapshot._change()
            elif snapshot._reason == "failing" == fix_reason:
                failing += 1
                snapshot._change()

        recorder.fix_all()
        if new:
            terminalreporter.write(f"defined values for {new} snapshots\n")
        if failing:
            terminalreporter.write(f"fixed {failing} snapshots\n")



from __future__ import annotations

import ast
import contextlib
import pathlib
from collections import defaultdict
from dataclasses import dataclass

import asttokens.util

try:
    from itertools import pairwise
except ImportError:
    from itertools import tee

    def pairwise(iterable):  # type: ignore
        a, b = tee(iterable)
        next(b, None)
        return zip(a, b)


# copied from pathlib to support python < 3.9
def is_relative_to(path, *other):
    """Return True if the path is relative to another path or False."""
    try:
        path.relative_to(*other)
        return True
    except ValueError:
        return False


@dataclass(order=True)
class SourcePosition:
    lineno: int
    col_offset: int


@dataclass(order=True)
class SourceRange:
    start: SourcePosition
    end: SourcePosition

    def __post_init__(self):
        if self.start > self.end:
            raise ValueError("range start should be lower then end")


@dataclass(order=True)
class Replacement:
    range: SourceRange
    text: str
    change_id: int = 0


def filename_of(obj):
    if hasattr(obj, "filename"):
        return obj.filename

    return None


def start_of(obj) -> SourcePosition:

    if isinstance(obj, asttokens.util.Token):
        return SourcePosition(lineno=obj.start[0], col_offset=obj.start[1])

    if isinstance(obj, SourcePosition):
        return obj

    if isinstance(obj, ast.AST):
        return SourcePosition(lineno=obj.lineno, col_offset=obj.col_offset)

    if isinstance(obj, SourceRange):
        return obj.start

    if isinstance(obj, tuple) and len(obj) == 2:
        return SourcePosition(lineno=obj[0], col_offset=obj[1])

    return None


def end_of(obj) -> SourcePosition:
    if isinstance(obj, asttokens.util.Token):
        return SourcePosition(lineno=obj.end[0], col_offset=obj.end[1])

    if isinstance(obj, ast.AST):
        return SourcePosition(lineno=obj.lineno, end_col_offset=obj.end_col_offset)

    if isinstance(obj, SourceRange):
        return obj.end

    return start_of(obj)


def range_of(obj, end=None):
    if isinstance(obj, tuple) and len(obj) == 2 and end is None:
        return SourceRange(start_of(obj[0]), end_of(obj[1]))

    if end is None:
        end = obj
    return SourceRange(start_of(obj), end_of(end))


class Change:  # ChangeSet
    _next_change_id = 0

    def __init__(self, change_recorder=None):
        self.change_recorder = change_recorder or ChangeRecorder.current
        if self.change_recorder is None:
            raise UsageError(
                "A change set needs a change recorder. Pass one as argument or use ChangeRecorder.activate"
            )

        self.change_recorder._changes.append(self)

        self.change_id = self._next_change_id
        self._tags = []
        type(self)._next_change_id += 1

    def set_tags(self, *tags):
        self._tags = tags

    def replace(self, node, new_contend, *, filename=None):
        if filename is None:
            filename = filename_of(node)

        if isinstance(new_contend, ast.AST):
            new_contend = ast.unparse(new_contend)

        if not isinstance(new_contend, str):
            new_contend = repr(new_contend)

        self._replace(
            filename,
            range_of(node),
            new_contend,
        )

    def delete(self, node, *, filename=None):
        self.replace(node, "", filename=filename)

    def insert(self, node, new_content, *, filename=None):
        self.replace(start_of(node), new_content, filename=filename)

    def _replace(self, filename, range, new_contend):

        if isinstance(new_contend, ast.AST):
            new_contend = ast.unparse(new_contend)

        self.change_recorder.get_source(filename).replacements.append(
            Replacement(range=range, text=new_contend, change_id=self.change_id)
        )


class SourceFile:
    def __init__(self, filename):
        self.replacements = []
        self.filename = filename

    def rewrite(self):
        new_code = self.new_code()

        if new_code is not None:
            with open(self.filename, "bw") as code:
                code.write(new_code.encode())

    def new_code(self):
        """returns the new file contend or None if there are no replacepents to
        apply."""
        replacements = list(self.replacements)
        replacements.sort()

        for r in replacements:
            assert r.range.start <= r.range.end

        # TODO check for overlapping replacements
        for lhs, rhs in pairwise(replacements):
            assert lhs.range.end <= rhs.range.start

        if not replacements:
            return

        with open(self.filename, newline="") as code:
            code = code.read()

        new_code = ""
        last_i = 0
        add_end = True

        for pos, i, c in code_stream(code):
            if replacements:
                r = replacements[0]
                if pos == r.range.start:
                    new_code += code[last_i:i] + r.text
                    add_end = False
                if pos == r.range.end:
                    last_i = i
                    add_end = True
                    replacements.pop(0)
            else:
                break

        if add_end:
            new_code += code[last_i:]

        return new_code

    def generate_patch(self, basedir):
        """yields lines of a pathfile."""

        filename = self.filename
        if is_relative_to(filename, basedir):
            filename = filename.relative_to(basedir)

        with open(self.filename, newline="") as code:
            old_code = code.read().splitlines(keepends=True)

        new_code = self.new_code().splitlines(keepends=True)

        import difflib

        yield from difflib.unified_diff(
            old_code, new_code, fromfile=str(filename), tofile=str(filename)
        )


def get_source_file(filename):
    filename = pathlib.Path(filename)
    return ChangeRecorder.current.get_source(filename)


def replace(node, new_contend):
    Change().replace(node, new_contend)


def insert_before(node, new_contend):
    if isinstance(new_contend, ast.AST):
        new_contend = ast.unparse(new_contend)
    new_contend += "\n"

    _replace(
        node.filename,
        (node.lineno, node.col_offset),
        (node.lineno, node.col_offset),
        new_contend,
    )


@contextlib.contextmanager
def code_change_disabled():
    with ChangeRecorder().activate():
        yield


class ChangeRecorder:
    current: ChangeRecorder | None = None

    def __init__(self):

        self._source_files = defaultdict(SourceFile)
        self._changes = []

    @contextlib.contextmanager
    def activate(self):
        old_recorder = ChangeRecorder.current
        ChangeRecorder.current = self
        yield self
        ChangeRecorder.current = old_recorder

    def get_source(self, filename):

        filename = pathlib.Path(filename)
        if filename not in self._source_files:
            self._source_files[filename] = SourceFile(filename)

        return self._source_files[filename]

    def change_set(self):
        return Change(self)

    def new_change(self):
        return Change(self)

    def changes(self):
        return list(self._changes)

    def num_fixes(self):
        changes = set()
        for file in self._source_files.values():
            changes.update(change.change_id for change in file.replacements)
        return len(changes)

    def fix_all(self, tags=()):
        for file in self._source_files.values():
            file.rewrite()

    def generate_patchfile(self, filename):
        with open(filename, "w") as patch:
            for line in self.generate_patch(filename.parent):
                patch.write(line)

    def generate_patch(self, basedir):
        for file in self._source_files.values():
            if is_relative_to(file.filename, basedir):
                yield from file.generate_patch(basedir)

    def dump(self):
        for file in self._source_files.values():
            print("file:", file.filename)
            for change in file.replacements:
                print("  change:", change)


global_recorder = ChangeRecorder()
ChangeRecorder.current = global_recorder


def code_stream(source):
    idx = 0
    p_line = 1
    p_col = 0
    while idx < len(source):
        c = source[idx]
        if c == "\r" and idx + 1 < len(source) and source[idx + 1] == "\n":
            # \r\n
            yield SourcePosition(p_line, p_col), idx, "\r\n"
            idx += 1
            p_line += 1
            p_col = 0
        elif c in "\r\n":
            # \r or \n
            yield SourcePosition(p_line, p_col), idx, c
            p_line += 1
            p_col = 0
        else:
            yield SourcePosition(p_line, p_col), idx, c
            p_col += 1
        idx += 1


def rewrite(filename=None):

    if filename is None:
        for file in ChangeRecorder.current._source_files.values:
            file.rewrite()
    else:
        if filename in ChangeRecorder.current._source_files:
            ChangeRecorder.current._source_files[filename].rewrite()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['inline_snapshot']

package_data = \
{'': ['*']}

install_requires = \
['asttokens>=2.0.5,<3.0.0', 'executing>=0.9.0,<0.10.0']

entry_points = \
{'pytest11': ['inline_snapshot = inline_snapshot.pytest_plugin']}

setup_kwargs = {
    'name': 'inline-snapshot',
    'version': '0.1.1',
    'description': 'compare test results with snapshots from previous test runs',
    'long_description': 'inline-snapshot\n======================\n\ncreate and update inline snapshots in your code.\n\nFeatures\n--------\n\n* records current values during [pytest](https://github.com/pytest-dev/pytest) run `--update-snapshots=new`.\n* values are stored in the source code and not in separate files.\n* values can be updated with `--update-snapshots=failing`.\n\n\nInstallation\n------------\n\nYou can install "inline-snapshot" via [pip](https://pypi.org/project/pip/) from [PyPI](https://pypi.org/project)::\n\n    $ pip install inline-snapshot\n\n\nUsage\n-----\n\nYou can use `snapshot()` instead of the value which you want to compare with.\n\n``` python\ndef something():\n    return 1548 * 18489\n\n\ndef test_something():\n    assert something() == snapshot()\n```\n\nYou can now run the tests and record the correct values.\n\n    $ pytest --update-snapshots=new\n\n``` python\ndef something():\n    return 1548 * 18489\n\n\ndef test_something():\n    assert something() == snapshot(28620972)  # snapshot gets recorded\n```\n\nYour tests will break if you change your code later.\nYou get normal pytest failure messages, because `snapshot(value)` just returns `value` during normal test runs.\n\n``` python\ndef something():\n    return (1548 * 18489) // 18  # changed implementation\n\n\ndef test_something():\n    assert something() == snapshot(28620972)  # this will fail now\n```\n\nMaybe that is correct and you should fix your code, or\nyour code is correct and you want to update your test results.\n\n    $ pytest --update-snapshots=failing\n\nPlease verify the new results. `git diff` will give you a good overview over all changed results.\nUse `pytest -k test_something --update-snapshots=failing` if you only want to change one test.\n\n``` python\ndef something():\n    return (1548 * 18489) // 18\n\n\ndef test_something():\n    assert something() == snapshot(1590054)\n```\n\nThe code is generated without any formatting.\nUse the formatter of your choice to make it look nice,\nor maybe use [darker](https://pypi.org/project/darker/) if you only want to format your changes.\n\n\nMore than just numbers\n----------------------\n\nRequirements:\n* `snapshot(value)` can only be used for `==` comparison\n* the values should be comparable with `==`\n* `repr(value)` should return valid python code\n\n\nYou can use almost any python datatype and also complex values like `datatime.date` (you have to import the right modules to match the `repr()` output).\n\n``` python\nfrom inline_snapshot import snapshot\nimport datetime\n\n\ndef something():\n    return {\n        "name": "hello",\n        "one number": 5,\n        "numbers": list(range(10)),\n        "sets": {1, 2, 15},\n        "datetime": datetime.date(1, 2, 22),\n        "complex stuff": 5j + 3,\n        "bytes": b"fglecg\\n\\x22",\n    }\n\n\ndef test_something():\n    assert something() == snapshot(\n        {\n            "name": "hello",\n            "one number": 5,\n            "numbers": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],\n            "sets": {1, 2, 15},\n            "datetime": datetime.date(1, 2, 22),\n            "complex stuff": (3 + 5j),\n            "bytes": b\'fglecg\\n"\',\n        }\n    )\n```\n\n\n`snapshot()` can also be used in loops.\n\n``` python\nfrom inline_snapshot import snapshot\n\n\ndef test_loop():\n    for name in ["Mia", "Ava", "Leo"]:\n        assert len(name) == snapshot(3)\n```\n\n… and more to come :grin:.\n\nContributing\n------------\nContributions are very welcome.\nTests can be run with [tox](https://tox.readthedocs.io/en/latest/).\nPlease use [pre-commit](https://pre-commit.com/) for your commits.\n\nLicense\n-------\n\nDistributed under the terms of the [MIT](http://opensource.org/licenses/MIT) license, "inline-snapshot" is free and open source software\n\n\nIssues\n------\n\nIf you encounter any problems, please [file an issue](https://github.com/15r10nk/pytest-inline-snapshot/issues) along with a detailed description.\n',
    'author': 'Frank Hoffmann',
    'author_email': '15r10nk@polarbit.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/15r10nk/inline-snapshots',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
